/*
 Illinois Open Source License

 University of Illinois/NCSA
 Open Source License

 Copyright © 2009,    University of Illinois.  All rights reserved.

 Developed by:
 Innovative Systems Lab
 National Center for Supercomputing Applications
 http://www.ncsa.uiuc.edu/AboutUs/Directorates/ISL.html

 Permission is hereby granted, free of charge, to any person obtaining a copy of this 
 software and associated documentation files (the "Software"), to deal with the Software 
 without restriction, including without limitation the rights to use, copy, modify, merge, 
 publish, distribute, sublicense, and/or sell copies of the Software, and to permit 
 persons to whom the Software is furnished to do so, subject to the following conditions:
 
 * Redistributions of source code must retain the above copyright notice, this list of 
   conditions and the following disclaimers.
 
 * Redistributions in binary form must reproduce the above copyright notice, this list of 
   conditions and the following disclaimers in the documentation and/or other materials 
   provided with the distribution.
 
 * Neither the names of Innovative Systems Lab and National Center for Supercomputing 
   Applications, nor the names of its contributors may be used to endorse or promote 
   products derived from this Software without specific prior written permission.
 
   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
   INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR 
   PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE 
   LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT 
   OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
   DEALINGS WITH THE SOFTWARE.
 */

#include "compute_task.h"
#include "spu_misc.h"
#include <sputask.h>

extern vector unsigned short large_table[];
#define BLOCK 1024
vector unsigned int index_array2[BLOCK/4];
vector unsigned short
compute_subtotal2(int iac)
{
    vector unsigned short zero_sv= (vector unsigned short){0,0,0,0,0,0,0,0};
    if (iac==0){
	return zero_sv;
    }
    int i;
    vector unsigned short subtotal = (vector unsigned short){0,0,0,0, 0,0,0,0};
    
    for (i=0;i < ROUNDDOWN_16(iac); i+=16){

#define URF 4 /*unroll factor*/	
	vector unsigned int * v = &index_array2[i];
	vector unsigned int t0[URF],t1[URF],t2[URF],t3[URF],t4[URF],t5[URF];
	int indx0[URF],indx1[URF],indx2[URF],indx3[URF];
	vector unsigned short entry0[URF], entry1[URF], entry2[URF], entry3[URF];
	vector unsigned short tsum0[URF], tsum1[URF];
	
#define ROTATE_AND_COMPOSE(idx) \
	t0[idx] = spu_rl(v[idx*4], 9); \
	t1[idx] = spu_rl(v[idx*4 + 1], 6); \
	t2[idx] = spu_rl(v[idx*4 + 2], 3); \
	t3[idx] = spu_or(t0[idx], t1[idx]); \
	t4[idx] = spu_or(t2[idx], v[idx*4+3]); \
	t5[idx] = spu_or(t3[idx], t4[idx]);  \
	indx0[idx] = spu_extract(t5[idx], 0); \
	indx1[idx] = spu_extract(t5[idx], 1); \
	indx2[idx] = spu_extract(t5[idx], 2); \
	indx3[idx] = spu_extract(t5[idx], 3); \
	entry0[idx] = large_table[indx0[idx]]; \
	entry1[idx] = large_table[indx1[idx]]; \
	entry2[idx] = large_table[indx2[idx]]; \
	entry3[idx] = large_table[indx3[idx]];	 \
	tsum0[idx] = spu_add(entry0[idx], entry1[idx]); \
	tsum1[idx] = spu_add(entry2[idx], entry3[idx]); \
	subtotal = spu_add(subtotal, tsum0[idx]); \
	subtotal = spu_add(subtotal, tsum1[idx]);
	
	ROTATE_AND_COMPOSE(0);
	ROTATE_AND_COMPOSE(1);
	ROTATE_AND_COMPOSE(2);
	ROTATE_AND_COMPOSE(3);

    }

    for (;i < ROUNDDOWN_4(iac); i+=4){
	
	vector unsigned int * v = &index_array2[i];
	vector unsigned int t0[URF],t1[URF],t2[URF],t3[URF],t4[URF],t5[URF];
	int indx0[URF],indx1[URF],indx2[URF],indx3[URF];
	vector unsigned short entry0[URF], entry1[URF], entry2[URF], entry3[URF];
	vector unsigned short tsum0[URF], tsum1[URF];
		
	ROTATE_AND_COMPOSE(0);

    }



    for (;i < iac; i++){
	
	int vl;
	for (vl =0;vl < 4;vl++){
	    vector unsigned short subtotal_addition[]={
		{1,0,0,0, 0,0,0,0},
		{0,1,0,0, 0,0,0,0},
		{0,0,1,0, 0,0,0,0},
		{0,0,0,1, 0,0,0,0},
		{0,0,0,0, 1,0,0,0},
		{0,0,0,0, 0,1,0,0},
		{0,0,0,0, 0,0,1,0},
		{0,0,0,0, 0,0,0,1}
	    };
	    
	    unsigned int indx = spu_extract(index_array2[i], vl);
	    subtotal = spu_add(subtotal, subtotal_addition[indx]);
	}
	
	
    }   
    
    return subtotal;
    
}

/* do a 4x4 float matrix transpose*/
void 
convert_back(vector float* src, vector float * dst)
{
    vector float v0, v1, v2, v3;
    vector unsigned char  trans_v[] = {
	{0,1,2,3,	16,17,18,19,	4,5,6,7,	20,21,22,23},
	{8,9,10,11,	24,25,26,27,	12,13,14,15,	28,29,30,31},
	{0,1,2,3,	4,5,6,7,	16,17,18,19,	20,21,22,23},
	{8,9,10,11,	12,13,14,15,	24,25,26,27,	28,29,30,31}
    };
    
    vector float * m = src;
    v0 = spu_shuffle(m[0], m[1], trans_v[0]);
    v1 = spu_shuffle(m[0], m[1], trans_v[1]);    
    v2 = spu_shuffle(m[2], m[3], trans_v[0]);
    v3 = spu_shuffle(m[2], m[3], trans_v[1]);

    dst[0] = spu_shuffle(v0, v2, trans_v[2]);
    dst[1] = spu_shuffle(v0, v2, trans_v[3]);
    dst[2] = spu_shuffle(v1, v3, trans_v[2]);
    dst[3] = spu_shuffle(v1, v3, trans_v[3]);
}
